@@ -79,15 +79,35 @@ static void print_usage(void)
 {
     struct proto *p;
     int i;
+    int res;
     char *prots = "";
 
     p = get_builtins();
-    for (i = 0; i < get_num_builtins(); i++)
-        asprintf(&prots, "%s\t[--%s <addr>]\n", prots, p[i].description);
+    for (i = 0; i < get_num_builtins(); i++) {
+        res = asprintf(&prots, "%s\t[--%s <addr>]\n", prots, p[i].description);
+        CHECK_RES_DIE(res, "asprintf");
+    }
 
     fprintf(stderr, USAGE_STRING, prots);
 }
 
+static void printcaps(void) {
+#ifdef LIBCAP
+    cap_t caps;
+    char* desc;
+    ssize_t len;
+
+    caps = cap_get_proc();
+
+    desc = cap_to_text(caps, &len);
+
+    fprintf(stderr, "capabilities: %s\n", desc);
+
+    cap_free(caps);
+    cap_free(desc);
+#endif
+}
+
 static void printsettings(void)
 {
     char buf[NI_MAXHOST];
@@ -254,7 +274,7 @@ static int config_protocols(config_t *config, struct proto **prots)
 static int config_parse(char *filename, struct addrinfo **listen, struct proto **prots)
 {
     config_t config;
-    long int timeout;
+    int timeout;
     const char* str;
 
     config_init(&config);
@@ -270,8 +290,9 @@ static int config_parse(char *filename, struct addrinfo **listen, struct proto *
     config_lookup_bool(&config, "inetd", &inetd);
     config_lookup_bool(&config, "foreground", &foreground);
     config_lookup_bool(&config, "numeric", &numeric);
+    config_lookup_bool(&config, "transparent", &transparent);
 
-    if (config_lookup_int(&config, "timeout", &timeout) == CONFIG_TRUE) {
+    if (config_lookup_int(&config, "timeout", (int *)&timeout) == CONFIG_TRUE) {
         probing_timeout = timeout;
     }
 
@@ -446,7 +467,7 @@ next_arg:
 
     set_protocol_list(prots);
 
-    if (!addr_listen) {
+    if (!addr_listen && !inetd) {
         fprintf(stderr, "No listening address specified; use at least one -p option\n");
         exit(1);
     }
@@ -504,9 +525,13 @@ int main(int argc, char *argv[])
    if (user_name)
        drop_privileges(user_name);
 
+
    /* Open syslog connection */
    setup_syslog(argv[0]);
 
+   if (verbose)
+       printcaps();
+
    main_loop(listen_sockets, num_addr_listen);
 
    return 0;
